import napari
import logging
from pathlib import Path

import napari.utils.notifications

import numpy as np
from PIL import Image
from scipy.ndimage import median_filter

from fibsem import acquire, constants
from fibsem.microscope import FibsemMicroscope, TescanMicroscope
from fibsem.structures import (BeamSettings, BeamType, FibsemDetectorSettings,
                               FibsemImage, ImageSettings, Point)
from fibsem.ui import utils as ui_utils

from fibsem.microscope import FibsemMicroscope, TescanMicroscope
from fibsem import constants, acquire

from fibsem.structures import BeamType, ImageSettings, FibsemImage, Point, FibsemDetectorSettings, BeamSettings
from fibsem.ui import utils as ui_utils 
from fibsem.ui import _stylesheets

from fibsem.ui.qtdesigner_files import ImageSettingsWidget
from fibsem.ui import _stylesheets
# from napari.qt.threading import thread_worker
   
from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal
import threading
from queue import Queue

import numpy as np
from fibsem import config as cfg
        

class FibsemImageSettingsWidget(ImageSettingsWidget.Ui_Form, QtWidgets.QWidget):
    picture_signal = pyqtSignal()
    viewer_update_signal = pyqtSignal()
    image_notification_signal = pyqtSignal(str)
    live_imaging_signal = pyqtSignal(dict)

    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        image_settings: ImageSettings = None,
        viewer: napari.Viewer = None,
        parent=None,
    ):
        super(FibsemImageSettingsWidget, self).__init__(parent=parent)
        self.setupUi(self)
        self.parent = parent
        self.microscope = microscope
        self.viewer = viewer
        self.eb_layer, self.ib_layer = None, None
        self.eb_image, self.ib_image = None, None

        self.eb_last = np.zeros(shape=(1024, 1536), dtype=np.uint8)
        self.ib_last = np.zeros(shape=(1024, 1536), dtype=np.uint8)

        self._features_layer = None
        self.stop_event = threading.Event()
        self.stop_event.set()
        self.image_queue = Queue()

        self._TESCAN = isinstance(self.microscope, TescanMicroscope)
        self.TAKING_IMAGES = False
        self._LIVE_IMAGING = False

        self.setup_connections()

        if image_settings is not None:
            # self.detector_settings = self.get_detector_settings()
            # self.beam_settings = self.get_beam_settings()
            self.image_settings = image_settings
            self.set_ui_from_settings(image_settings = image_settings, beam_type = BeamType.ELECTRON)
        self.update_detector_ui() # TODO: can this be removed?

        # register initial images
        self.update_viewer(np.zeros(shape=(1024, 1536), dtype=np.uint8), BeamType.ION.name)
        self.update_viewer(np.zeros(shape=(1024, 1536), dtype=np.uint8), BeamType.ELECTRON.name)
        self.live_imaging_signal.connect(self.live_update)
    
    def setup_connections(self):

        # set ui elements
        self.selected_beam.addItems([beam.name for beam in BeamType])

        # buttons
        self.pushButton_take_image.clicked.connect(lambda: self.take_image(None))
        self.pushButton_take_all_images.clicked.connect(self.take_reference_images)
        self.pushButton_live_imaging.clicked.connect(self.live_imaging)
                
        # feature flags
        self.pushButton_live_imaging.setVisible(cfg._LIVE_IMAGING_ENABLED)
        
        # image / beam settings
        self.selected_beam.currentIndexChanged.connect(self.update_detector_ui)
        self.checkBox_image_save_image.toggled.connect(self.update_ui_saving_settings)
        self.button_set_beam_settings.clicked.connect(self.apply_beam_settings)
        
        # detector
        self.set_detector_button.clicked.connect(self.apply_detector_settings)
        self.detector_contrast_slider.valueChanged.connect(self.update_labels)
        self.detector_brightness_slider.valueChanged.connect(self.update_labels)
        
        # util
        self.ion_ruler_checkBox.toggled.connect(self.update_ruler)
        self.scalebar_checkbox.toggled.connect(self.update_ui_tools)
        self.crosshair_checkbox.toggled.connect(self.update_ui_tools)
        
        # signals
        self.image_notification_signal.connect(self.update_imaging_ui)
        
        # advanced settings
        self.checkBox_advanced_settings.stateChanged.connect(self.toggle_mode)
        self.toggle_mode()


        # set ui stylesheets
        self.pushButton_take_image.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.pushButton_take_all_images.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.pushButton_live_imaging.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.set_detector_button.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.button_set_beam_settings.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)

        if self._TESCAN:

            self.label_stigmation.hide()
            self.stigmation_x.hide()
            self.stigmation_y.hide()
            self.stigmation_x.setEnabled(False)
            self.stigmation_y.setEnabled(False)
            available_presets = self.microscope.get_available_values("presets")
            self.comboBox_presets.addItems(available_presets)   
            self.comboBox_presets.currentTextChanged.connect(self.update_presets)
        else:
            self.comboBox_presets.hide()
            self.label_presets.hide()
  
    def toggle_mode(self):
        """Toggle the visibility of the advanced settings"""
        advanced_mode = self.checkBox_advanced_settings.isChecked()

        self.label_detector_type.setVisible(advanced_mode)
        self.detector_type_combobox.setVisible(advanced_mode)
        self.label_detector_mode.setVisible(advanced_mode)
        self.detector_mode_combobox.setVisible(advanced_mode)
        self.label_stigmation.setVisible(advanced_mode)
        self.stigmation_x.setVisible(advanced_mode)
        self.stigmation_y.setVisible(advanced_mode)
        self.shift_x.setVisible(advanced_mode)
        self.shift_y.setVisible(advanced_mode)
        self.label_shift.setVisible(advanced_mode)
        self.beam_voltage.setVisible(advanced_mode)
        self.label_beam_voltage.setVisible(advanced_mode)
        self.label_beam_scan_rotation.setVisible(advanced_mode)
        self.spinBox_beam_scan_rotation.setVisible(advanced_mode)
        self.checkBox_image_use_autocontrast.setVisible(advanced_mode)
        self.checkBox_image_use_autogamma.setVisible(advanced_mode)
            


    def update_presets(self):
        beam_type = BeamType[self.selected_beam.currentText()]
        self.microscope.set("preset", self.comboBox_presets.currentText(), beam_type)
    def check_point_image(self,point):
            
            if point[1] >= 0 and point[1] <= self.eb_layer.data.shape[1]:
                return True
            else:
                return False

    def update_ruler(self):

        if self.ion_ruler_checkBox.isChecked():
            self.ion_ruler_label.setText("Ruler: is on")

            # create initial ruler

            data = [[500,500],[500,1000]]
            p1,p2 = data[0],data[1]


            hfw_scale = self.eb_image.metadata.pixel_size.x if self.check_point_image(p1) else self.ib_image.metadata.pixel_size.x

            midpoint = [np.mean([p1[0],p2[0]]),np.mean([p1[1],p2[1]])]
            dist_um = 500 * hfw_scale*constants.SI_TO_MICRO
            text = {
                "string": [f"{dist_um:.2f} um"],
                "color": "white"
            }

            # creating initial layers 

            self._features_layer = self.viewer.add_points(data, size=20, face_color='lime', edge_color='white', name='ruler')
            self.viewer.add_shapes(data, shape_type='line', edge_color='lime', name='ruler_line',edge_width=5)
            self.viewer.add_points(midpoint,text=text, size=20, face_color='transparent', edge_color='transparent', name='ruler_value')
            self._features_layer.mode = 'select'


            self.viewer.layers.selection.active = self._features_layer
            self._features_layer.mouse_drag_callbacks.append(self.update_ruler_points)


        else:
            self.ion_ruler_label.setText("Ruler: is off")
            self.viewer.layers.remove(self._features_layer)
            self.viewer.layers.remove('ruler_line')
            self.viewer.layers.remove('ruler_value')
            self._features_layer = None
            self.viewer.layers.selection.active = self.eb_layer




    def update_ruler_points(self,layer, event):
        
        dragged = False
        yield

        while event.type == 'mouse_move':


            if self._features_layer.selected_data is not None:
                data = self._features_layer.data


                p1 = data[0]
                p2 = data[1]

                dist_pix = np.linalg.norm(p1-p2)
                
                midpoint = [(np.mean([p1[0],p2[0]])),(np.mean([p1[1],p2[1]]))]
                
                self.viewer.layers['ruler_line'].data = [p1,p2]
                self.viewer.layers['ruler_value'].data = midpoint
                
                hfw_scale = self.eb_image.metadata.pixel_size.x if self.check_point_image(p1) else self.ib_image.metadata.pixel_size.x

                dist_um = dist_pix * hfw_scale*constants.SI_TO_MICRO

                text = {
                "string": [f"{dist_um:.2f} um"],
                "color": "white"
                }

                self.viewer.layers['ruler_value'].text = text
                dist_dx = abs(p2[1]-p1[1]) * self.image_settings.hfw/self.image_settings.resolution[0]*constants.SI_TO_MICRO
                dist_dy = abs(p2[0]-p1[0]) * self.image_settings.hfw/self.image_settings.resolution[0]*constants.SI_TO_MICRO


                self.ion_ruler_label.setText(f"Ruler: {dist_um:.2f} um  dx: {dist_dx:.2f} um  dy: {dist_dy:.2f} um")
                self.viewer.layers.selection.active = self._features_layer
                self.viewer.layers['ruler_line'].refresh()

                
                dragged = True
                yield
            



    def update_labels(self):
        self.detector_contrast_label.setText(f"{self.detector_contrast_slider.value()}%")
        self.detector_brightness_label.setText(f"{self.detector_brightness_slider.value()}%")

    def apply_detector_settings(self):
        
        # read settings from ui
        beam =  BeamType[self.selected_beam.currentText()]
        self.get_settings_from_ui()

        # set detector settings
        self.microscope.set_detector_settings(self.detector_settings, beam_type=beam)
        
        # logging
        logging.debug({"msg": "apply_detector_settings", "detector_settings": self.detector_settings.to_dict()})

        # notifications
        napari.utils.notifications.show_info("Detector Settings Updated")

    def apply_beam_settings(self):
        beam = BeamType[self.selected_beam.currentText()]
        self.get_settings_from_ui()

        # set beam settings
        self.microscope.set_beam_settings(self.beam_settings)

        # logging 
        logging.debug({"msg": "apply_beam_settings", "beam_settings": self.beam_settings.to_dict()})

        # QUERY: why is this here?
        self.set_ui_from_settings(self.image_settings,beam)
        
        # notifications
        napari.utils.notifications.show_info("Beam Settings Updated")

    def get_settings_from_ui(self):

        self.image_settings = ImageSettings(
            resolution=[self.spinBox_resolution_x.value(), self.spinBox_resolution_y.value()],
            dwell_time=self.doubleSpinBox_image_dwell_time.value() * constants.MICRO_TO_SI,
            hfw=self.doubleSpinBox_image_hfw.value() * constants.MICRO_TO_SI,
            beam_type=BeamType[self.selected_beam.currentText()],
            autocontrast=self.checkBox_image_use_autocontrast.isChecked(),
            autogamma=self.checkBox_image_use_autogamma.isChecked(),
            save=self.checkBox_image_save_image.isChecked(),
            path=Path(self.lineEdit_image_path.text()),
            filename=self.lineEdit_image_label.text()
            
        )

        self.detector_settings = FibsemDetectorSettings(
            type=self.detector_type_combobox.currentText(),
            mode=self.detector_mode_combobox.currentText(),
            brightness=self.detector_brightness_slider.value()*constants.FROM_PERCENTAGES,
            contrast=self.detector_contrast_slider.value()*constants.FROM_PERCENTAGES,
        )

        self.beam_settings = BeamSettings(
            beam_type=BeamType[self.selected_beam.currentText()],
            working_distance=self.working_distance.value()*constants.MILLI_TO_SI,
            beam_current=self.beam_current.value()*constants.PICO_TO_SI,
            voltage=self.beam_voltage.value()*constants.KILO_TO_SI,
            hfw = self.doubleSpinBox_image_hfw.value() * constants.MICRO_TO_SI,
            resolution=[self.spinBox_resolution_x.value(), self.spinBox_resolution_y.value()],
            dwell_time=self.doubleSpinBox_image_dwell_time.value() * constants.MICRO_TO_SI,
            stigmation = Point(self.stigmation_x.value(), self.stigmation_y.value()),
            shift = Point(self.shift_x.value() * constants.MICRO_TO_SI, self.shift_y.value()*constants.MICRO_TO_SI),
            scan_rotation = np.deg2rad(self.spinBox_beam_scan_rotation.value())
        )
        return self.image_settings, self.detector_settings, self.beam_settings

    def set_ui_from_settings(self, image_settings: ImageSettings, 
                             beam_type: BeamType, 
                             beam_settings: BeamSettings=None, 
                             detector_settings: FibsemDetectorSettings=None ):
        """Update the ui from the image, beam and detector settings"""

        # disconnect beam type combobox
        self.selected_beam.currentIndexChanged.disconnect()
        self.selected_beam.setCurrentText(beam_type.name)
        self.selected_beam.currentIndexChanged.connect(self.update_detector_ui)
        
        if beam_settings is None:
            beam_settings = self.microscope.get_beam_settings(beam_type)
        if detector_settings is None:
            detector_settings = self.microscope.get_detector_settings(beam_type)

        # imaging settings
        self.spinBox_resolution_x.setValue(image_settings.resolution[0])
        self.spinBox_resolution_y.setValue(image_settings.resolution[1])

        self.doubleSpinBox_image_dwell_time.setValue(image_settings.dwell_time * constants.SI_TO_MICRO)
        self.doubleSpinBox_image_hfw.setValue(image_settings.hfw * constants.SI_TO_MICRO)

        self.checkBox_image_use_autocontrast.setChecked(image_settings.autocontrast)
        self.checkBox_image_use_autogamma.setChecked(image_settings.autogamma)
        self.checkBox_image_save_image.setChecked(image_settings.save)
        self.lineEdit_image_path.setText(str(image_settings.path))
        self.lineEdit_image_label.setText(image_settings.filename)

        # detector settings
        self.detector_type_combobox.setCurrentText(detector_settings.type)
        self.detector_mode_combobox.setCurrentText(detector_settings.mode)
        self.detector_contrast_slider.setValue(int(detector_settings.contrast*100))
        self.detector_brightness_slider.setValue(int(detector_settings.brightness*100))
        
        # beam settings
        self.beam_current.setValue(beam_settings.beam_current*constants.SI_TO_PICO)
        self.beam_voltage.setValue(beam_settings.voltage*constants.SI_TO_KILO)
        self.spinBox_beam_scan_rotation.setValue(np.rad2deg(beam_settings.scan_rotation))
        
        self.working_distance.setValue(beam_settings.working_distance*constants.METRE_TO_MILLIMETRE)
        self.shift_x.setValue(beam_settings.shift.x * constants.SI_TO_MICRO)
        self.shift_y.setValue(beam_settings.shift.y * constants.SI_TO_MICRO)
        self.stigmation_x.setValue(beam_settings.stigmation.x)
        self.stigmation_y.setValue(beam_settings.stigmation.y)
        
        self.update_ui_saving_settings()

    def update_ui_saving_settings(self):
        """Toggle the visibility of the imaging saving settings"""
        self.label_image_save_path.setVisible(self.checkBox_image_save_image.isChecked())
        self.lineEdit_image_path.setVisible(self.checkBox_image_save_image.isChecked())
        self.label_image_label.setVisible(self.checkBox_image_save_image.isChecked())
        self.lineEdit_image_label.setVisible(self.checkBox_image_save_image.isChecked())
        
  
    def update_detector_ui(self):
        """Update the detector ui based on currently selected beam"""
        beam_type = BeamType[self.selected_beam.currentText()]

        _is_ion = bool(beam_type is BeamType.ION)
        _is_tescan = isinstance(self.microscope, TescanMicroscope)
        
        self.comboBox_presets.setVisible(_is_ion and _is_tescan)
        self.label_presets.setVisible(_is_ion and _is_tescan)


        available_detector_types = self.microscope.get_available_values("detector_type", beam_type=beam_type)
        self.detector_type_combobox.clear()
        self.detector_type_combobox.addItems(available_detector_types)
        self.detector_type_combobox.setCurrentText(self.microscope.get("detector_type", beam_type=beam_type))
        
        available_detector_modes = self.microscope.get_available_values("detector_mode", beam_type=beam_type)
        if available_detector_modes is None: 
            available_detector_modes = ["N/A"]
        self.detector_mode_combobox.clear()
        self.detector_mode_combobox.addItems(available_detector_modes)
        self.detector_mode_combobox.setCurrentText(self.microscope.get("detector_mode", beam_type=beam_type))

        self.set_ui_from_settings(self.image_settings, beam_type)

    def live_imaging(self):
        if self.stop_event.is_set():
            self._toggle_interactions(False)
            self.pushButton_live_imaging.setEnabled(True)
            self.pushButton_live_imaging.setText("Stop live imaging")
            self.parent.movement_widget.checkBox_movement_acquire_electron.setChecked(False)
            self.parent.movement_widget.checkBox_movement_acquire_ion.setChecked(False)
            self.pushButton_live_imaging.setStyleSheet("background-color: orange")

            self.stop_event.clear()
            self.image_queue.queue.clear()
            image_settings = self.get_settings_from_ui()[0]
            image_settings.autocontrast = False
            from copy import deepcopy
            _thread = threading.Thread(
                target=self.microscope.live_imaging,
                args=(deepcopy(image_settings), self.image_queue, self.stop_event),
            )
            _thread.start()
            sleep_time = image_settings.dwell_time*image_settings.resolution[0]*image_settings.resolution[1]
            worker = self.microscope.consume_image_queue(parent_ui = self, sleep = sleep_time)
            worker.returned.connect(self.update_live_finished)
            import time
            time.sleep(1)
            worker.start()  
            self._LIVE_IMAGING = True

        else:
            self._LIVE_IMAGING = False
            self.stop_event.set()
            self.pushButton_live_imaging.setStyleSheet("""
                    QPushButton {
                        background-color: green;
                        color: white;
                    }
                    QPushButton:hover {
                        background-color: gray;
                    }
                    QPushButton:pressed { """
                )
            self.parent.movement_widget.checkBox_movement_acquire_electron.setChecked(False)
            self.parent.movement_widget.checkBox_movement_acquire_ion.setChecked(False)


    def update_live_finished(self):
        self.pushButton_live_imaging.setText("Live imaging")
        self.pushButton_take_all_images.setEnabled(True)
        self.pushButton_take_image.setEnabled(True)
        self._toggle_interactions(True)

    def live_update(self, dict):
        arr = dict["image"].data
        name = BeamType[self.selected_beam.currentText()].name

        try:
            self.viewer.layers[name].data = arr
        except:    
            layer = self.viewer.add_image(arr, name = name)

        if name == BeamType.ELECTRON.name:
            self.eb_image = dict["image"]
        if name == BeamType.ION.name:
            self.ib_image = dict["image"]    
        
    def take_image(self, beam_type: BeamType = None):
        self.TAKING_IMAGES = True
        worker = self.take_image_worker(beam_type)
        worker.finished.connect(self.imaging_finished)
        worker.start()

    def update_imaging_ui(self, msg: str):
        logging.info(msg)
        napari.utils.notifications.notification_manager.records.clear()
        napari.utils.notifications.show_info(msg)

    def imaging_finished(self):
        if self.ib_image is not None:
            self.update_viewer(self.ib_image.data, BeamType.ION.name)
        if self.eb_image is not None:
            self.update_viewer(self.eb_image.data, BeamType.ELECTRON.name)
        self._toggle_interactions(True)
        self.TAKING_IMAGES = False

    def _toggle_interactions(self, enable: bool, caller: str = None, imaging: bool = False):
        self.pushButton_take_image.setEnabled(enable)
        self.pushButton_live_imaging.setEnabled(enable)
        self.pushButton_take_all_images.setEnabled(enable)
        self.set_detector_button.setEnabled(enable)
        self.button_set_beam_settings.setEnabled(enable)
        self.parent.movement_widget._toggle_interactions(enable, caller="ui")
        if caller != "milling":
            self.parent.milling_widget._toggle_interactions(enable, caller="ui")
        if enable:
            self.pushButton_take_all_images.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.pushButton_take_image.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.pushButton_live_imaging.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
            self.set_detector_button.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
            self.button_set_beam_settings.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
            self.pushButton_take_image.setText("Acquire Image")
            self.pushButton_take_all_images.setText("Acquire All Images")
        elif imaging:
            self.pushButton_take_all_images.setStyleSheet(_stylesheets._ORANGE_PUSHBUTTON_STYLE)
            self.pushButton_take_image.setStyleSheet(_stylesheets._ORANGE_PUSHBUTTON_STYLE)
            self.pushButton_take_image.setText("Acquiring Images...")
            self.pushButton_take_all_images.setText("Acquiring Images...")
            self.set_detector_button.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.button_set_beam_settings.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_live_imaging.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
        else:
            self.pushButton_take_all_images.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_take_image.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.set_detector_button.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.button_set_beam_settings.setStyleSheet(_stylesheets._DISABLED_PUSHBUTTON_STYLE)
            self.pushButton_take_image.setText("Acquire Image")
            self.pushButton_take_all_images.setText("Acquire All Images")

    from napari.qt.threading import thread_worker
                
    @thread_worker
    def take_image_worker(self, beam_type: BeamType = None):
        self._toggle_interactions(enable=False, imaging=True)
        self.image_settings = self.get_settings_from_ui()[0]
        self.image_notification_signal.emit("Acquiring Image...")
        if beam_type is not None:
            self.image_settings.beam_type = beam_type

        arr =  acquire.new_image(self.microscope, self.image_settings)
        name = f"{self.image_settings.beam_type.name}"

        if self.image_settings.beam_type == BeamType.ELECTRON:
            self.eb_image = arr
        if self.image_settings.beam_type == BeamType.ION:
            self.ib_image = arr
        
        self.picture_signal.emit()

        logging.debug({"msg": "take_image_worker", "image_settings": self.image_settings.to_dict()})


    def take_reference_images(self):
        self.TAKING_IMAGES = True
        worker = self.take_reference_images_worker()
        worker.finished.connect(self.imaging_finished)
        worker.start()

    @thread_worker
    def take_reference_images_worker(self):
        self._toggle_interactions(enable=False,imaging=True)
        self.image_settings = self.get_settings_from_ui()[0]
        self.image_notification_signal.emit("Acquiring Images...")
        self.eb_image, self.ib_image = acquire.take_reference_images(self.microscope, self.image_settings)

        self.picture_signal.emit()

        logging.debug({"msg": "take_reference_images_worker", "image_settings": self.image_settings.to_dict()})

    def update_ui_tools(self):
        """Redraw the ui tools (scalebar, crosshair)"""
        self.update_viewer(self.eb_last, BeamType.ELECTRON.name)
        self.update_viewer(self.ib_last, BeamType.ION.name)

    def update_viewer(self, arr: np.ndarray, name: str, _set_ui: bool = False):
        """Update the viewer with the given image array"""

        if name == BeamType.ELECTRON.name:
            self.eb_last = arr
        if name == BeamType.ION.name:
            self.ib_last = arr

        # median filter for display
        arr = median_filter(arr, size=3)
       
        try:
            self.viewer.layers[name].data = arr
        except:    
            layer = self.viewer.add_image(arr, name = name)
        

        layer = self.viewer.layers[name]
        if self.eb_layer is None and name == BeamType.ELECTRON.name:
            self.eb_layer = layer
        if self.ib_layer is None and name == BeamType.ION.name:
            self.ib_layer = layer
        

        # centre the camera
        if self.eb_layer:
            self.viewer.camera.center = [
                0.0,
                self.eb_layer.data.shape[0] / 2,
                self.eb_layer.data.shape[1],
            ]
            self.viewer.camera.zoom = 0.45

        if self.ib_layer:
            translation = (
                self.viewer.layers["ELECTRON"].data.shape[1]
                if self.eb_layer
                else arr.shape[1]
            )
            self.ib_layer.translate = [0.0, translation]       

        if self.eb_layer:
            points = np.array([[-20, 200], [-20, self.eb_layer.data.shape[1] + 150]])
            string = ["ELECTRON BEAM", "ION BEAM"]
            text = {
                "string": string,
                "color": "white"
            }

            try:
                self.viewer.layers['label'].data = points
            except:    
                self.viewer.add_points(
                points,
                name="label",
                text=text,
                size=20,
                edge_width=7,
                edge_width_is_relative=False,
                edge_color='transparent',
                face_color='transparent',
                )   


        # draw scalebar and crosshair
        if self.eb_image is not None and self.ib_image is not None:
            ui_utils._draw_scalebar(viewer=self.viewer,eb_image= self.eb_image,ib_image= self.ib_image,is_checked=self.scalebar_checkbox.isChecked())
            ui_utils._draw_crosshair(viewer=self.viewer,eb_image= self.eb_image,ib_image= self.ib_image,is_checked=self.crosshair_checkbox.isChecked()) 
            
        
        # set ui from image metadata
        if _set_ui:
            if name == BeamType.ELECTRON.name:
                self.image_settings = self.eb_image.metadata.image_settings
                beam_settings = self.eb_image.metadata.microscope_state.electron_beam
                detector_settings = self.eb_image.metadata.microscope_state.electron_detector
                beam_type = BeamType.ELECTRON
            if name == BeamType.ION.name:
                self.image_settings = self.ib_image.metadata.image_settings
                beam_settings = self.ib_image.metadata.microscope_state.ion_beam
                detector_settings = self.ib_image.metadata.microscope_state.ion_detector
                beam_type = BeamType.ION            
        else:
            beam_type = BeamType[self.selected_beam.currentText()]
            beam_settings, detector_settings = None, None    
    
        self.set_ui_from_settings(image_settings = self.image_settings, beam_type= beam_type, 
            beam_settings=beam_settings, detector_settings=detector_settings)  
            
        # set the active layer to the electron beam (for movement)
        if self.eb_layer:
            self.viewer.layers.selection.active = self.eb_layer

        self.viewer_update_signal.emit()

    def get_data_from_coord(self, coords: tuple) -> tuple:
        # check inside image dimensions, (y, x)
        eb_shape = self.eb_image.data.shape[0], self.eb_image.data.shape[1]
        ib_shape = self.ib_image.data.shape[0], self.ib_image.data.shape[1] + self.eb_image.data.shape[1]

        if (coords[0] > 0 and coords[0] < eb_shape[0]) and (
            coords[1] > 0 and coords[1] < eb_shape[1]
        ):
            image = self.eb_image
            beam_type = BeamType.ELECTRON

        elif (coords[0] > 0 and coords[0] < ib_shape[0]) and (
            coords[1] > eb_shape[0] and coords[1] < ib_shape[1]
        ):
            image = self.ib_image
            coords = (coords[0], coords[1] - ib_shape[1] // 2)
            beam_type = BeamType.ION
        else:
            beam_type, image = None, None

        # logging
        logging.debug( {"msg": "get_data_from_coord", "coords": coords, "beam_type": beam_type})

        return coords, beam_type, image
    
    def closeEvent(self, event):
        self.viewer.layers.clear()
        event.accept()

    def clear_viewer(self):
        self.viewer.layers.clear()
        self.eb_layer = None
        self.ib_layer = None

    def _set_active_layer(self):
        if self.eb_layer:
            self.viewer.layers.selection.active = self.eb_layer

def main():

    image_settings = ImageSettings(resolution=[1536, 1024], dwell_time=1e-6, hfw=150e-6, 
    autocontrast=True, beam_type=BeamType.ION, 
    save=True, filename="my_label", path="path/to/save", 
    autogamma=True)

    viewer = napari.Viewer(ndisplay=2)
    image_settings_ui = FibsemImageSettingsWidget(image_settings=image_settings)
    viewer.window.add_dock_widget(
        image_settings_ui, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()



METADATA_VERSION = "v3"

# sputtering rates, from microscope application files
MILLING_SPUTTER_RATE = {
    20e-12: 6.85e-3,  # 30kv
    0.2e-9: 6.578e-2,  # 30kv
    0.74e-9: 3.349e-1,  # 30kv
    0.89e-9: 3.920e-1,  # 20kv
    2.0e-9: 9.549e-1,  # 30kv
    2.4e-9: 1.309,  # 20kv
    6.2e-9: 2.907,  # 20kv
    7.6e-9: 3.041,  # 30kv
    28.0e-9: 1.18e1,  # 30 kv
}

SUPPORTED_COORDINATE_SYSTEMS = [
    "RAW",
    "SPECIMEN",
    "STAGE",
    "Raw",
    "raw",
    "specimen",
    "Specimen",
    "Stage",
    "stage",
]


REFERENCE_HFW_WIDE = 2750e-6
REFERENCE_HFW_LOW = 900e-6
REFERENCE_HFW_MEDIUM = 400e-6
REFERENCE_HFW_HIGH = 150e-6
REFERENCE_HFW_SUPER = 80e-6
REFERENCE_HFW_ULTRA = 50e-6

REFERENCE_RES_SQUARE = [1024, 1024]
REFERENCE_RES_LOW = [768, 512]
REFERENCE_RES_MEDIUM = [1536, 1024]
REFERENCE_RES_HIGH = [3072, 2048]
REFERENCE_RES_SUPER = [6144, 4096]

MILL_HFW_THRESHOLD = 0.005  # 0.5% of the image


import os
import fibsem

BASE_PATH = os.path.dirname(
    fibsem.__path__[0]
)  # TODO: figure out a more stable way to do this
CONFIG_PATH = os.path.join(BASE_PATH, "fibsem", "config")
PROTOCOL_PATH = os.path.join(CONFIG_PATH, "protocol.yaml")
LOG_PATH = os.path.join(BASE_PATH, "fibsem", "log")
DATA_PATH = os.path.join(BASE_PATH, "fibsem", "log", "data")
DATA_ML_PATH: str = os.path.join(BASE_PATH, "fibsem", "log", "data", "ml")
DATA_CC_PATH: str = os.path.join(BASE_PATH, "fibsem", "log", "data", "crosscorrelation")
DATA_TILE_PATH: str = os.path.join(DATA_PATH, "tile")
POSITION_PATH = os.path.join(CONFIG_PATH, "positions.yaml")
MODELS_PATH = os.path.join(BASE_PATH, "fibsem", "segmentation", "models")
MICROSCOPE_CONFIGURATION_PATH = os.path.join(
    CONFIG_PATH, "microscope-configuration.yaml"
)



os.makedirs(LOG_PATH, exist_ok=True)
os.makedirs(DATA_PATH, exist_ok=True)
os.makedirs(DATA_ML_PATH, exist_ok=True)
os.makedirs(DATA_CC_PATH, exist_ok=True)
os.makedirs(DATA_TILE_PATH, exist_ok=True)

DATABASE_PATH = os.path.join(BASE_PATH, "fibsem", "db", "fibsem.db")
os.makedirs(os.path.dirname(DATABASE_PATH), exist_ok=True)
import yaml


def load_yaml(fname):
    """
    Load a YAML file and return its contents as a dictionary.

    Args:
        fname (str): The path to the YAML file to be loaded.

    Returns:
        dict: A dictionary containing the contents of the YAML file.

    Raises:
        IOError: If the file cannot be opened or read.
        yaml.YAMLError: If the file is not valid YAML.
    """
    with open(fname, "r") as f:
        config = yaml.safe_load(f)

    return config


__SUPPORTED_MANUFACTURERS__ = ["Thermo", "Tescan", "Demo"]
__DEFAULT_MANUFACTURER__ = "Thermo"
__DEFAULT_IP_ADDRESS__ = "192.168.0.1"
__SUPPORTED_PLASMA_GASES__ = ["Argon", "Oxygen", "Nitrogen", "Xenon"]


def get_default_user_config() -> dict:
    """Return the default configuration."""
    return {
        "name":                           "default-configuration",       # a descriptive name for your configuration 
        "ip_address":                     __DEFAULT_IP_ADDRESS__,        # the ip address of the microscope PC
        "manufacturer":                   __DEFAULT_MANUFACTURER__,      # the microscope manufactuer, Thermo, Tescan or Demo                       
        "rotation-reference":             0,                             # the reference rotation value (rotation when loading)  [degrees]
        "shuttle-pre-tilt":               35,                            # the pre-tilt of the shuttle                           [degrees]
        "electron-beam-eucentric-height": 7.0e-3,                        # the eucentric height of the electron beam             [metres]
        "ion-beam-eucentric-height":      16.5e-3,                       # the eucentric height of the ion beam                  [metres]
    }


# user configurations -> move to fibsem.db eventually
USER_CONFIGURATIONS_PATH = os.path.join(CONFIG_PATH, "user-configurations.yaml")
USER_CONFIGURATIONS_YAML = load_yaml(USER_CONFIGURATIONS_PATH)
USER_CONFIGURATIONS = USER_CONFIGURATIONS_YAML["configurations"]
DEFAULT_CONFIGURATION_NAME = USER_CONFIGURATIONS_YAML["default"]
DEFAULT_CONFIGURATION_PATH = USER_CONFIGURATIONS[DEFAULT_CONFIGURATION_NAME]["path"]


if DEFAULT_CONFIGURATION_PATH is None:
    USER_CONFIGURATIONS[DEFAULT_CONFIGURATION_NAME][
        "path"
    ] = MICROSCOPE_CONFIGURATION_PATH
    DEFAULT_CONFIGURATION_PATH = MICROSCOPE_CONFIGURATION_PATH

if not os.path.exists(DEFAULT_CONFIGURATION_PATH):
    DEFAULT_CONFIGURATION_NAME = "default-configuration"
    USER_CONFIGURATIONS[DEFAULT_CONFIGURATION_NAME][
        "path"
    ] = MICROSCOPE_CONFIGURATION_PATH
    DEFAULT_CONFIGURATION_PATH = MICROSCOPE_CONFIGURATION_PATH
        
print(f"Default configuration: {DEFAULT_CONFIGURATION_NAME}")
print(f"Default configuration path: {DEFAULT_CONFIGURATION_PATH}")

def add_configuration(configuration_name: str, path: str):
    """Add a new configuration to the user configurations file."""
    if configuration_name in USER_CONFIGURATIONS:
        raise ValueError(f"Configuration name '{configuration_name}' already exists.")

    USER_CONFIGURATIONS[configuration_name] = {"path": path}
    USER_CONFIGURATIONS_YAML["configurations"] = USER_CONFIGURATIONS
    with open(USER_CONFIGURATIONS_PATH, "w") as f:
        yaml.dump(USER_CONFIGURATIONS_YAML, f)


def remove_configuration(configuration_name: str):
    """Remove a configuration from the user configurations file."""
    if configuration_name not in USER_CONFIGURATIONS:
        raise ValueError(f"Configuration name '{configuration_name}' does not exist.")

    del USER_CONFIGURATIONS[configuration_name]
    USER_CONFIGURATIONS_YAML["configurations"] = USER_CONFIGURATIONS
    with open(USER_CONFIGURATIONS_PATH, "w") as f:
        yaml.dump(USER_CONFIGURATIONS_YAML, f)


def set_default_configuration(configuration_name: str):
    """Set the default configuration in the user configurations file."""
    if configuration_name not in USER_CONFIGURATIONS:
        raise ValueError(f"Configuration name '{configuration_name}' does not exist.")

    USER_CONFIGURATIONS_YAML["default"] = configuration_name
    with open(USER_CONFIGURATIONS_PATH, "w") as f:
        yaml.dump(USER_CONFIGURATIONS_YAML, f)


# default configuration values
DEFAULT_CONFIGURATION_VALUES = {
    "Thermo": {
        "ion-column-tilt": 52,
        "electron-column-tilt": 0,
    },
    "Tescan": {
        "ion-column-tilt": 55,
        "electron-column-tilt": 0,
    },
    "Demo": {
        "ion-column-tilt": 52,
        "electron-column-tilt": 0,
    },
}


# machine learning
HUGGINFACE_REPO = "patrickcleeve/autolamella"
__DEFAULT_CHECKPOINT__ = "autolamella-mega-20240107.pt"




# feature flags

_LIVE_IMAGING_ENABLED = False
_MINIMAP_VISUALISATION = False
_MINIMAP_MOVE_WITH_TRANSLATION = False
_MINIMAP_ACQUIRE_AFTER_MOVEMENT = False
_APPLY_CONFIGURATION_ENABLED = True



# tescan manipulator

TESCAN_MANIPULATOR_CALIBRATION_PATH = os.path.join(CONFIG_PATH, "tescan_manipulator.yaml")

def load_tescan_manipulator_calibration() -> dict:
    """Load the tescan manipulator calibration"""
    config = utils.load_yaml(cfg.TESCAN_MANIPULATOR_CALIBRATION_PATH)
    return config

def save_tescan_manipulator_calibration(config: dict) -> None:
    """Save the tescan manipulator calibration"""
    utils.save_yaml(cfg.TESCAN_MANIPULATOR_CALIBRATION_PATH, config)
    return None


import argparse
import os

import pandas as pd
import yaml

from fibsem.detection import evaluation

def main(config: dict):

    os.makedirs(config["save_path"], exist_ok=True)

    if config["run_eval"]:
        df_eval = evaluation.run_evaluation_v2(path=config["data_path"],
                                             image_path = config["images_path"], 
                                                checkpoints=config["checkpoints"], 
                                                plot=config["show_det_plot"],
                                                save=config["save_det_plot"], 
                                                save_path=config["save_path"])
    else:
        df_eval = pd.read_csv(os.path.join(config["save_path"], "eval.csv"))

    if config["plot_eval"]:
        category_orders = {"checkpoint": df_eval["checkpoint"].unique().tolist(), 
                    "feature": sorted(df_eval["feature"].unique().tolist())}
        evaluation.plot_evaluation_data(df=df_eval, 
                                        category_orders=category_orders,
                                        thresholds=config["thresholds"], 
                                        show=config["show_eval_plot"], 
                                        save=config["save_eval_plot"],
                                        save_path=config["save_path"])



if __name__ == "__main__":
# command line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config",
        help="the directory containing the config file to use",
        dest="config",
        action="store",
        default=os.path.join(os.path.join(os.path.dirname(__file__), "config.yml")),
    )
    args = parser.parse_args()
    config_dir = args.config

    # NOTE: Setup your config.yml file
    with open(config_dir, "r") as f:
        config = yaml.safe_load(f)

    main(config=config)


import logging
import os
import yaml
from pathlib import Path
import napari
import napari.utils.notifications
from PyQt5 import QtWidgets
from PyQt5.QtGui import QPixmap, QImage
from copy import deepcopy
from fibsem import constants
from fibsem.microscope import FibsemMicroscope
from fibsem.structures import FibsemStagePosition
from fibsem.ui.FibsemImageSettingsWidget import FibsemImageSettingsWidget
from fibsem.ui.qtdesigner_files import FibsemPositionsWidget, FibsemMovementWidget
from fibsem.ui.utils import _get_save_file_ui, _get_file_ui


class FibsemPositionsWidget(FibsemPositionsWidget.Ui_Form, QtWidgets.QWidget):
    def __init__(
        self,
        microscope: FibsemMicroscope = None,
        movement_widget: FibsemMovementWidget = None,
        image_widget: FibsemImageSettingsWidget = None,
        parent=None,
    ):
        super(FibsemPositionsWidget, self).__init__(parent=parent)
        self.setupUi(self)

        self.microscope = microscope
        self.movement_widget = movement_widget
        self.image_widget = image_widget
        self.setup_connections()

        self.positions = []

    def setup_connections(self):
        self.comboBox_positions.currentIndexChanged.connect(self.select_position)
        self.pushButton_save_position.clicked.connect(self.add_position)
        self.pushButton_remove_position.clicked.connect(self.delete_position)
        self.pushButton_go_to.clicked.connect(self.go_to_position)
        self.pushButton_export.clicked.connect(self.export_positions)
        self.pushButton_import.clicked.connect(self.import_positions)
        self.movement_widget.move_signal.connect(self.minimap)

    def select_position(self):
        if self.comboBox_positions.currentIndex() != -1:
            position = self.positions[self.comboBox_positions.currentIndex()]
            self.label_current_position.setText(f"x={position.x*constants.METRE_TO_MILLIMETRE:.3f}, y={position.y*constants.METRE_TO_MILLIMETRE:.3f}, z={position.z*constants.METRE_TO_MILLIMETRE:.3f}, r={position.r*constants.RADIANS_TO_DEGREES:.1f}, t={position.t*constants.RADIANS_TO_DEGREES:.1f}")
            self.minimap()

    def add_position(self):
        position = self.microscope.get_stage_position()
        name = self.lineEdit_position_name.text()
        if name == "":
            napari.utils.notifications.show_warning("Please enter a name for the position")
            return
        position.name = name
        self.positions.append(position)
        self.comboBox_positions.addItem(name)
        self.comboBox_positions.setCurrentIndex(self.comboBox_positions.count() - 1)
        self.lineEdit_position_name.setText("")
        logging.info(f"Added position {name}")

    def delete_position(self):
        del self.positions[self.comboBox_positions.currentIndex()]
        name = self.comboBox_positions.currentIndex()
        self.comboBox_positions.removeItem(self.comboBox_positions.currentIndex())
        logging.info(f"Removed position {name}")

    def go_to_position(self):
        self.microscope.move_stage_absolute(self.positions[self.comboBox_positions.currentIndex()])
        self.movement_widget.update_ui()
        self.image_widget.take_reference_images()
        logging.info(f"Moved to position {self.comboBox_positions.currentIndex()}")

    def export_positions(self):
        protocol_path = _get_save_file_ui(msg="Select or create file")
        if protocol_path == '':
            return
        dict_position = []
        for position in self.positions:
            dict_position.append(position.to_dict())
        with open(os.path.join(Path(protocol_path).with_suffix(".yaml")), "w") as f:
            yaml.safe_dump(dict_position, f, indent=4, default_flow_style=False)

        logging.info("Positions saved to file")


    def import_positions(self):
        protocol_path = _get_file_ui(msg="Select or create file")
        if protocol_path == '':
            napari.utils.notifications.show_info("No file selected, positions not loaded")
            return
        with open(protocol_path, "r") as f:
            dict_positions = yaml.safe_load(f)
        for dict_position in dict_positions:
            position = FibsemStagePosition.from_dict(dict_position)
            self.positions.append(position)
            self.comboBox_positions.addItem(position.name)

    def minimap(self):
        x = []
        y = []
        labels = []
        pil_image = None
        current_position = self.microscope.get_stage_position()
        x.append(deepcopy(current_position.x)*constants.SI_TO_MICRO)
        y.append(deepcopy(current_position.y)*constants.SI_TO_MICRO)
        labels.append("Current Position")
        for position in self.positions:
            x.append(deepcopy(position.x)*constants.SI_TO_MICRO)
            y.append(deepcopy(position.y)*constants.SI_TO_MICRO)
            labels.append(deepcopy(position.name))
        import pandas as pd
        df = pd.DataFrame({'x': x, 'y': y, 'labels': labels})
        import plotly.express as px
        import plotly.io as pio
        fig = px.scatter(df, color="labels", labels={'color': 'Position'}, title="Positions (um)", x = 'x', y = 'y')

        # miminum graph zoom 
        if len(self.positions) > 0:
            flag = False
            for i in range(len(self.positions)):
                if (x[i] - x[0]) < 150 or (y[i] - y[0]) < 150:
                    flag = True
            if flag:
                range_x = [min(x) - 75, max(x) + 75]
                range_y = [min(y) - 75, max(y) + 75]
                fig.update_layout(
                    xaxis=dict(range=range_x),
                    yaxis=dict(range=range_y)
                )

        image_from_plot = fig.to_image(format="png", engine="kaleido")

        from PIL import Image
        import io
        pil_image = Image.open(io.BytesIO(image_from_plot))
        # Convert the PIL image to a QImage
        image_qt = QImage(pil_image.tobytes(), pil_image.width, pil_image.height, QImage.Format_RGBA8888)
        # Convert the QImage to a QPixmap
        qpixmap = QPixmap.fromImage(image_qt)
        self.label_minimap.setPixmap(qpixmap)

def main():

    viewer = napari.Viewer(ndisplay=2)
    movement_widget = FibsemPositionsWidget()
    viewer.window.add_dock_widget(
        movement_widget, area="right", add_vertical_stretch=False
    )
    napari.run()


if __name__ == "__main__":
    main()



import logging
import os
import sys
import glob

import fibsem
import napari
import napari.utils.notifications
import numpy as np
import tifffile as tff
from PIL import Image
from fibsem.ui.qtdesigner_files import FibsemLabellingUI
from PyQt5 import QtWidgets
from fibsem.segmentation.config import CLASS_COLORS, CLASS_LABELS, convert_color_names_to_rgb, CLASS_CONFIG_PATH
from fibsem.segmentation import utils as seg_utils

from fibsem.ui.FibsemSegmentationModelWidget import FibsemSegmentationModelWidget
from fibsem.ui import _stylesheets
from fibsem.ui.utils import _get_directory_ui,_get_file_ui

from napari.layers import Points
from typing import Any, Generator, Optional

# setup a basic logger
logging.basicConfig(stream=sys.stdout, level=logging.INFO)

_DEBUG = False

# new line char in html
NL = "<br>"

START_INSTRUCTIONS = f"""
<h3>Instructions</h3>
Welcome to the OpenFIBSEM Labelling Tool.{NL}
This tool is designed to assist with labelling 2D images for segmentation.{NL}

<h4>Getting Started</h4>
Select the folder containing the image data.{NL}
Select the folder to save the labels to.{NL}
Select a custom colormap [Optional].{NL}

"""
READY_INSTRUCTIONS = f"""
<h4>Ready to Label</h4>
Use the paintbrush to label the image.{NL}
Use the eraser to remove labels.{NL}
Press R/E to go to the next/previous image.{NL}
Press D/F to change the selected class label.{NL}{NL}

<h4>Saving</h4>
The image will be saved when you move to the next image.{NL}
You can also save the image by pressing S.{NL}{NL}

"""

MODEL_NOT_LOADED_INSTRUCTIONS = f"""
<h4>Model Assisted Labelling</h4>
You can use a model to assist with labelling.{NL}
Select the model to use for segmentation.{NL}
The model will be used to generate a mask for the image.{NL}
You are then able to use the paintbrush to correct the mask.{NL}{NL}
"""

MODEL_LOADED_INSTRUCTIONS = f"""
<h4>Model Assisted Labelling</h4>
The Segmentation Model will assist with labelling images.{NL}
The model will be used to generate a mask for the image.{NL}
You are then able to use the paintbrush to correct the mask.{NL}{NL}
"""

SAM_MODEL_LOADED_INSTRUCTIONS = f"""
<h4>Segment Anything Model</h4>
The SAM Model will assist with labelling images.{NL}
Left Click to add a point to the mask.{NL}
Right Click to remove a point from the mask.{NL}
Press C to confirm the SAM mask.{NL}
Press X to clear the SAM mask.{NL}{NL}


"""

INSTRUCTIONS = {
    "START": START_INSTRUCTIONS,
    "READY": READY_INSTRUCTIONS,
    "MODEL_NOT_LOADED": MODEL_NOT_LOADED_INSTRUCTIONS,
    "MODEL_LOADED": MODEL_LOADED_INSTRUCTIONS,
    "SAM_MODEL_LOADED": SAM_MODEL_LOADED_INSTRUCTIONS,
}

CONFIGURATION = {
    "IMAGES": {
        "FILE_EXT": ".tif",
        "SUPPORTED_FILE_EXT": [".tif", ".png", ".jpg", ".jpeg"]
    },
    "LABELS": {
        "COLOR_MAP": CLASS_COLORS,
        "LABEL_MAP": CLASS_LABELS,
    },
    "SAVE": {
        "SAVE_RGB": True,
    },
    "UI": {
        "INSTRUCTIONS": INSTRUCTIONS,
        "LOAD_DATA_BUTTON_COLOR": _stylesheets._BLUE_PUSHBUTTON_STYLE,
        "LOAD_MODEL_BUTTON_COLOR": _stylesheets._BLUE_PUSHBUTTON_STYLE,
        "CONFIRM_BUTTON_COLOR": _stylesheets._GREEN_PUSHBUTTON_STYLE,
        "CLEAR_BUTTON_COLOR": _stylesheets._RED_PUSHBUTTON_STYLE,
        "IMAGE_OPACITY": 0.7,
        "MASK_OPACITY": 0.3,
    },
    "TOOLTIPS": {
        "AUTOSAVE": "Automatically save the image when moving to the next image",
        "SAVE_RGB": "Save the RGB mask when saving the image. This is in additional to the class mask, and is only for visualisation purposes.",
    }

}

# ref (other sam labelling tools):
# https://github.com/JoOkuma/napari-segment-anything (Apache License 2.0)
# https://github.com/MIC-DKFZ/napari-sam (Apache License 2.0)
class FibsemLabellingUI(FibsemLabellingUI.Ui_Dialog, QtWidgets.QDialog):
    def __init__(
        self,
        viewer: napari.Viewer,
        parent=None,
    ):
        super(FibsemLabellingUI, self).__init__(parent=parent)
        self.setupUi(self)
        self.viewer = viewer
        self.last_idx = 0
        self.model = None
        self.model_type = None
        self.model_assist = False
        self.DATA_LOADED = False

        self.img_layer = None
        self.mask_layer = None
        
        # sam
        self.sam_mask_layer = None
        self.sam_pts_layer = None   
        self.logits = None          # sam logits
        self.image = None           # sam image

        self.setup_connections()

        if _DEBUG:
            self.lineEdit_data_path.setText("/home/patrick/github/data/autolamella-paper/model-development/train/serial-liftout/train")
            self.lineEdit_labels_path.setText("/home/patrick/github/fibsem/fibsem/log/labels2")
            self.model_widget.lineEdit_checkpoint.setText("autolamella-serial-liftout-20240107.pt")
    
    def setup_connections(self):
        self.pushButton_load_data.clicked.connect(self.load_data)

        self.model_widget = FibsemSegmentationModelWidget()
        
        # save path buttons

        self.pushButton_data_path.clicked.connect(self.select_filepath)
        self.pushButton_labels_path.clicked.connect(self.select_filepath)
        self.model_widget.checkpoint_seg_button.clicked.connect(self.select_filepath)
        self.pushButton_data_config.clicked.connect(self.select_filepath)
        self.lineEdit_data_config.setText(CLASS_CONFIG_PATH)
        self.comboBox_data_file_ext.addItems(CONFIGURATION["IMAGES"]["SUPPORTED_FILE_EXT"])

        self.tabWidget.addTab(self.model_widget, "Model")
        
        # set tab 3 invisible (until fixed)
        self.tabWidget.setTabEnabled(2, False)
        self.tabWidget.setTabVisible(2, False)        

        self.model_widget.model_loaded.connect(self.load_model)

        self.checkBox_model_assist.stateChanged.connect(self.toggle_model_assist)

        self.viewer.bind_key("R", self.next_image)
        self.viewer.bind_key("E", self.previous_image)
        self.viewer.bind_key("S", self.save_image)
        self.viewer.bind_key("D", self.decrement_class_index)
        self.viewer.bind_key("F", self.increment_class_index)

        # SAM
        self.pushButton_model_confirm.setVisible(False)
        self.pushButton_model_clear.setVisible(False)

        # style
        self.pushButton_load_data.setStyleSheet(CONFIGURATION["UI"]["LOAD_DATA_BUTTON_COLOR"])
        self.model_widget.pushButton_load_model.setStyleSheet(CONFIGURATION["UI"]["LOAD_MODEL_BUTTON_COLOR"])
        self.pushButton_model_confirm.setStyleSheet(CONFIGURATION["UI"]["CONFIRM_BUTTON_COLOR"])
        self.pushButton_model_clear.setStyleSheet(CONFIGURATION["UI"]["CLEAR_BUTTON_COLOR"])

        # tooltips
        self.checkBox_autosave.setToolTip(CONFIGURATION["TOOLTIPS"]["AUTOSAVE"])
        self.checkBox_save_rgb.setToolTip(CONFIGURATION["TOOLTIPS"]["SAVE_RGB"])

        self.update_instructions()

    def select_filepath(self):

        if self.sender() == self.pushButton_data_path:
            path = _get_directory_ui(msg="Select Image Data Directory")
            if path is not None and path != "":
                self.lineEdit_data_path.setText(path)
        elif self.sender() == self.pushButton_labels_path:
            path = _get_directory_ui(msg="Select Labels Directory")
            if path is not None and path != "":
                self.lineEdit_labels_path.setText(path)

        elif self.sender() == self.model_widget.checkpoint_seg_button:
            path = _get_file_ui(msg="Select Checkpoint File", _filter=None)
            if path is not None and path != "":
                self.model_widget.lineEdit_checkpoint.setText(path)
        elif self.sender() == self.pushButton_data_config:
            path = _get_file_ui(msg="Select Configuration File", path=CLASS_CONFIG_PATH, _filter="*.yaml")
            if path is not None and path != "":
                self.lineEdit_data_config.setText(path)

                import yaml
                with open(path) as f:
                    CLASS_CONFIG = yaml.load(f, Loader=yaml.FullLoader) 

                CONFIGURATION["LABELS"]["COLOR_MAP"] = CLASS_CONFIG["CLASS_COLORS"]
                CONFIGURATION["LABELS"]["LABEL_MAP"] = CLASS_CONFIG["CLASS_LABELS"]


    def load_data(self):
        # read raw data
        data_path = self.lineEdit_data_path.text()
        labels_path = self.lineEdit_labels_path.text()
        
        # create required directories
        os.makedirs(labels_path, exist_ok=True)
        
        # check if save and raw are the same 
        if data_path == labels_path:
            napari.utils.notifications.show_error(f"Save and Raw directories cannot be the same")
            return
            
        # get filenames
        CONFIGURATION["IMAGES"]["FILE_EXT"] = self.comboBox_data_file_ext.currentText()
        FILE_EXT = CONFIGURATION["IMAGES"]["FILE_EXT"]
        filenames = sorted(glob.glob(os.path.join(data_path, f"*{FILE_EXT}*")))
        if len(filenames) == 0:
            napari.utils.notifications.show_error(f"No images found in {data_path}")
            return

        # assign data
        self.data_path = data_path
        self.labels_path = labels_path
        self.filenames = filenames
        self.DATA_LOADED = True

        # remove existing layers
        try:
            self.viewer.layers.remove(self.img_layer)
            self.viewer.layers.remove(self.mask_layer)
            self.viewer.layers.remove(self.sam_pts_layer)
            self.viewer.layers.remove(self.sam_mask_layer)
        except:
            pass

        # load filenames as single layer
        self.img_layer = self.viewer.open(filenames, name="Image", stack=True)[0]
        self.img_layer.opacity = CONFIGURATION["UI"]["IMAGE_OPACITY"]
        self.img_layer.blending="additive"

        arr = np.zeros_like(np.asarray(self.img_layer.data[0]))
        self.mask_layer = self.viewer.add_labels(
            arr,
            name="Mask",
            opacity=CONFIGURATION["UI"]["MASK_OPACITY"],
            blending="additive",
            color=CONFIGURATION["LABELS"]["COLOR_MAP"],
        )

        self.viewer.dims.events.current_step.connect(self.update_image)
        self.update_viewer_to_image(0)

        self.update_ui_elements()

        self.update_instructions()

        # TODO: fix this event, it just doesn't work?
        # self.mask_layer.events.data.connect(self.save_image) 

    def update_ui_elements(self):
        # update ui elements on data loaded
        n_labels = len(CONFIGURATION["LABELS"]["LABEL_MAP"])
        ldict = CONFIGURATION['LABELS']['LABEL_MAP']
        cdict = CONFIGURATION['LABELS']['COLOR_MAP']
        label_map = [f"{i:02d} - {ldict.get(i, 'Unspecified')} ({cdict.get(i, 'Unspecified')})" for i in range(n_labels)]
        self.comboBox_model_class_index.clear()
        self.comboBox_model_class_index.addItems(label_map)

    def previous_image(self , _: Optional[Any] = None) -> None:
        idx = int(self.viewer.dims.point[0])
        idx -= 1
        idx = np.clip(idx, 0, len(self.filenames) - 1)
        self.update_viewer_to_image(idx)

    def next_image(self, _: Optional[Any] = None) -> None:
        idx = int(self.viewer.dims.point[0])
        idx += 1
        idx = np.clip(idx, 0, len(self.filenames) - 1)
        self.update_viewer_to_image(idx)

    def save_image(self, _: Optional[Any] = None) -> None:

        # save current image
        idx = self.last_idx
        fname = self.filenames[idx]
        FILE_EXT = CONFIGURATION["IMAGES"]["FILE_EXT"]
        bname = os.path.basename(fname).removesuffix(FILE_EXT)

        # only resave the labels...
        label = np.asarray(self.mask_layer.data).astype(np.uint8)

        im = Image.fromarray(label) # TODO: replace with tifffile?
        im.save(os.path.join(self.labels_path, f"{bname}{FILE_EXT}"))
        
        logging.info(f"Saving mask to {os.path.basename(fname)}")

        if CONFIGURATION["SAVE"]["SAVE_RGB"] and self.checkBox_save_rgb.isChecked():
            
            colormap = CONFIGURATION["LABELS"]["COLOR_MAP"]
            colormap_rgb = convert_color_names_to_rgb(colormap)

            rgb = seg_utils.decode_segmap_v2(label, colormap_rgb)
            rgb = Image.fromarray(rgb)

            RGB_PATH = os.path.join(self.labels_path, "rgb")
            os.makedirs(RGB_PATH, exist_ok=True)
            rgb.save(os.path.join(RGB_PATH, f"{bname}.png"))

            logging.info(f"Saving RGB mask to {os.path.join(RGB_PATH, f'{bname}.png')}")

    
    def update_viewer_to_image(self, idx: int):
        self.viewer.dims.set_point(0, idx)

    def update_image(self):
        
        idx = int(self.viewer.dims.point[0])
        logging.info(f"UPDATING IMAGE: LAST IDX: {self.last_idx}, CURRENT IDX {idx}")
        
        # save previous image
        if self.last_idx != idx:
            if self.checkBox_autosave.isChecked():
                self.save_image()
            self.last_idx = idx # update last index
        
        # update mask layer
        self.mask_layer.data = self.get_label_image()

        # set active layers
        if self.model is not None:
            if self.model_type == "SegmentAnythingModel":
                self.model.is_image_set = False # TODO: implement this, so we don't have to recompute the image embedding each time
                self.logits = None
                self.image = None
        self.set_active_layers()

    def set_active_layers(self):
        model_assist = self.checkBox_model_assist.isChecked()
        if model_assist and self.sam_pts_layer is not None:
            self.viewer.layers.selection.active = self.sam_pts_layer
            self.sam_pts_layer.mode = "add"
        else:
            self.viewer.layers.selection.active = self.mask_layer
            self.mask_layer.mode = "paint"

    def update_instructions(self):
        # display instructions
        msg=""

        if not self.DATA_LOADED:
            msg = INSTRUCTIONS["START"]
        else: 
            msg = INSTRUCTIONS["READY"]

            if self.model is None: 
                msg += INSTRUCTIONS["MODEL_NOT_LOADED"]

            elif self.model is not None:

                if self.model_type == "SegmentAnythingModel":
                    msg += INSTRUCTIONS["SAM_MODEL_LOADED"]
                else:
                    msg += INSTRUCTIONS["MODEL_LOADED"]

        self.label_instructions.setText(msg)

    def get_label_image(self) -> np.ndarray:

        idx = int(self.viewer.dims.point[0])
        fname = self.filenames[idx]
        image = np.asarray(self.img_layer.data[idx]) # req for lazy load

        logging.info(f"IDX: {idx}, Loading image from {os.path.basename(fname)}")

        if os.path.basename(fname) in os.listdir(os.path.join(self.labels_path)):
            label_fname = os.path.join(self.labels_path, os.path.basename(fname))
            label_image = Image.open(label_fname).convert("L")
            label_image = np.array(label_image, dtype=np.uint8)

            msg = f"Loaded label image from {os.path.basename(label_fname)}"

        elif self.model_assist and self.model is not None:
            if self.model_type == "SegmentationModel":
                label_image = self.model.inference(image, rgb=False)
                if label_image.ndim == 3:
                    label_image = label_image[0]
                msg = f"Generated label image using {os.path.basename(self.model.checkpoint)}"
            if self.model_type == "SegmentAnythingModel":
                label_image = np.zeros_like(image)

                msg = f"No label image found, using SAM model to generate mask."

        else:
            label_image = np.zeros_like(image)

            msg = "No label image found, and no model loaded"

        napari.utils.notifications.show_info(msg)

        return np.asarray(label_image).astype(np.uint8)

    ####### MODEL
    def toggle_model_assist(self):
        self.model_assist = self.checkBox_model_assist.isChecked()

        if self.model is None:
            self.label_model_info.setText(f"Please load a model.")
        if self.DATA_LOADED:
            self.update_image()

    def load_model(self):
        self.model = self.model_widget.model
        self.model_type =  self.model_widget.model_type
        self.label_model_info.setText(f"Model: {os.path.basename(self.model.checkpoint)}")

        # specific layers for SAM model
        if self.model_type == "SegmentAnythingModel":
            self._add_sam_pts_layer()
            
            self.pushButton_model_confirm.clicked.connect(self._confirm_model_class)
            self.pushButton_model_clear.clicked.connect(self._clear_sam_data)
            self.pushButton_model_confirm.setVisible(True)

            self.pushButton_model_clear.setVisible(True)

            self.viewer.bind_key("C", self._confirm_model_class, overwrite=True)
            self.viewer.bind_key("X", self._clear_sam_data, overwrite=True)

        napari.utils.notifications.show_info(
            f"Loaded {self.model_type}: {os.path.basename(self.model.checkpoint)}"
        )

        self.update_instructions()

        if self.DATA_LOADED:
            self.update_image()

    ####### SAM
    def _add_sam_pts_layer(self):
        # add SAM points layer
        self.sam_pts_layer = self.viewer.add_points(name="SAM Points")
        self.sam_pts_layer.events.data.connect(self._update_sam_mask)
        self.sam_pts_layer.mouse_drag_callbacks.append(self._mouse_button_modifier)
        self.sam_pts_layer.mode = "add"

    def _mouse_button_modifier(self, _: Points, event) -> None:
        self.sam_pts_layer.selected_data = []
        if event.button == 1:
            self.sam_pts_layer.current_face_color = "blue"
            self.sam_pts_layer.current_properties = {"prompt": 1}
        else:
            self.sam_pts_layer.current_face_color = "red"
            self.sam_pts_layer.current_properties = {"prompt": 0} # 1 = pos, 0 = neg, -1 = background

    def _update_sam_mask(self, _: Optional[Any] = None) -> None:

        points = None
        points_labels = None
        # get the prompt points from sam_pts layer
        if len(self.sam_pts_layer.data) > 0:
            points = [np.flip(self.sam_pts_layer.data, axis=-1).tolist()]
            points_labels = [self.sam_pts_layer.properties["prompt"].tolist()]

        if self.image is None:
            idx = int(self.viewer.dims.point[0])
            image = np.asarray(self.img_layer.data[idx]) # req for lazy load
            self.image = Image.fromarray(image).convert("RGB")

        # TODO: dont' recompute image embedding each time
        mask, score, self.logits = self.model.predict(self.image, 
                                                points, 
                                                points_labels, 
                                                input_masks=self.logits,
                                                multimask_output=False)

        # add layer for mask
        try:
            self.sam_mask_layer.data = mask
        except:
            self.sam_mask_layer = self.viewer.add_labels(
                mask,
                name="SAM Mask",
                opacity=0.7,
                blending="additive",
                color={0: "black", 1: "white"},
            )

        self.set_sam_active()

    def set_sam_active(self):
        try:
            # set sam points layer active
            self.viewer.layers.selection.active = self.sam_pts_layer
            self.sam_pts_layer.mode = "add"
        except:
            pass

    def _confirm_model_class(self, _: Optional[Any] = None) -> None:
        if self.sam_mask_layer is None:
            return

        # take the current sam mask and add it to the Mask
        sam_mask = self.sam_mask_layer.data

        # get the current class index
        cidx = self.comboBox_model_class_index.currentIndex()

        # set all values of sam mask to class index
        sam_mask[sam_mask > 0] = cidx

        # add to, and update mask layer
        mask = self.mask_layer.data
        mask[sam_mask > 0] = sam_mask[sam_mask > 0]
        self.mask_layer.data = mask

        self._clear_sam_data()

    def increment_class_index(self, _: Optional[Any] = None) -> None:
        self.change_class_index(1)

    def decrement_class_index(self, _: Optional[Any] = None) -> None:
        self.change_class_index(-1)

    def change_class_index(self, val: int = 0) -> None:
        # get the current class index
        cidx = self.comboBox_model_class_index.currentIndex()
        cidx += val

        # clip to 0,number of classes
        n_labels = len(CONFIGURATION["LABELS"]["LABEL_MAP"])
        cidx = np.clip(cidx, 0, n_labels - 1)

        self.comboBox_model_class_index.setCurrentIndex(cidx)
        self.viewer.status = f"Current Class: {cidx}"

        # set the label index for the mask layer
        if self.mask_layer is not None:
            self.mask_layer.selected_label = cidx

    def _clear_sam_data(self, _: Optional[Any] = None) -> None:
        # clear points
        # TODO: find the proper way to do this, not doing this way causses issues with face colour
        self.viewer.layers.remove(self.sam_pts_layer)
        self._add_sam_pts_layer()

        # clear mask
        if self.sam_mask_layer is not None:
            self.sam_mask_layer.data = np.zeros_like(self.sam_mask_layer.data)

    def closeEvent(self, event):
        # try to save the current image on close
        try:
            self.save_image()
        except:
            pass
        event.accept()


def main():
    viewer = napari.Viewer(ndisplay=2)
    fibsem_labelling_ui = FibsemLabellingUI(viewer=viewer)
    viewer.window.add_dock_widget(
        fibsem_labelling_ui,
        area="right",
        add_vertical_stretch=True,
        name=f"OpenFIBSEM v{fibsem.__version__} - Image Labelling",

    )
    napari.run()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemMultiChemWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.7
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(256, 240)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.label_hfw = QtWidgets.QLabel(Form)
        self.label_hfw.setObjectName("label_hfw")
        self.gridLayout.addWidget(self.label_hfw, 5, 0, 1, 1)
        self.label_gas = QtWidgets.QLabel(Form)
        self.label_gas.setObjectName("label_gas")
        self.gridLayout.addWidget(self.label_gas, 4, 0, 1, 1)
        self.label_title = QtWidgets.QLabel(Form)
        self.label_title.setObjectName("label_title")
        self.gridLayout.addWidget(self.label_title, 0, 0, 1, 2)
        self.comboBox_beam_type = QtWidgets.QComboBox(Form)
        self.comboBox_beam_type.setObjectName("comboBox_beam_type")
        self.gridLayout.addWidget(self.comboBox_beam_type, 1, 1, 1, 1)
        self.pushButton_run_sputtering = QtWidgets.QPushButton(Form)
        self.pushButton_run_sputtering.setObjectName("pushButton_run_sputtering")
        self.gridLayout.addWidget(self.pushButton_run_sputtering, 8, 0, 1, 2)
        self.label_position = QtWidgets.QLabel(Form)
        self.label_position.setObjectName("label_position")
        self.gridLayout.addWidget(self.label_position, 3, 0, 1, 1)
        self.label_application_file = QtWidgets.QLabel(Form)
        self.label_application_file.setObjectName("label_application_file")
        self.gridLayout.addWidget(self.label_application_file, 2, 0, 1, 1)
        self.label_beam_type = QtWidgets.QLabel(Form)
        self.label_beam_type.setObjectName("label_beam_type")
        self.gridLayout.addWidget(self.label_beam_type, 1, 0, 1, 1)
        self.doubleSpinBox_hfw = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_hfw.setMaximum(2700.0)
        self.doubleSpinBox_hfw.setProperty("value", 150.0)
        self.doubleSpinBox_hfw.setObjectName("doubleSpinBox_hfw")
        self.gridLayout.addWidget(self.doubleSpinBox_hfw, 5, 1, 1, 1)
        self.doubleSpinBox_length = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_length.setObjectName("doubleSpinBox_length")
        self.gridLayout.addWidget(self.doubleSpinBox_length, 6, 1, 1, 1)
        self.label_length = QtWidgets.QLabel(Form)
        self.label_length.setObjectName("label_length")
        self.gridLayout.addWidget(self.label_length, 6, 0, 1, 1)
        self.comboBox_position = QtWidgets.QComboBox(Form)
        self.comboBox_position.setObjectName("comboBox_position")
        self.gridLayout.addWidget(self.comboBox_position, 3, 1, 1, 1)
        self.comboBox_application_file = QtWidgets.QComboBox(Form)
        self.comboBox_application_file.setObjectName("comboBox_application_file")
        self.gridLayout.addWidget(self.comboBox_application_file, 2, 1, 1, 1)
        self.comboBox_gas = QtWidgets.QComboBox(Form)
        self.comboBox_gas.setObjectName("comboBox_gas")
        self.gridLayout.addWidget(self.comboBox_gas, 4, 1, 1, 1)
        self.label_time = QtWidgets.QLabel(Form)
        self.label_time.setObjectName("label_time")
        self.gridLayout.addWidget(self.label_time, 7, 0, 1, 1)
        self.doubleSpinBox_time = QtWidgets.QDoubleSpinBox(Form)
        self.doubleSpinBox_time.setProperty("value", 30.0)
        self.doubleSpinBox_time.setObjectName("doubleSpinBox_time")
        self.gridLayout.addWidget(self.doubleSpinBox_time, 7, 1, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_hfw.setText(_translate("Form", "HFW (um)"))
        self.label_gas.setText(_translate("Form", "Gas"))
        self.label_title.setText(_translate("Form", "MultiChem"))
        self.pushButton_run_sputtering.setText(_translate("Form", "Run Sputtering"))
        self.label_position.setText(_translate("Form", "Position"))
        self.label_application_file.setText(_translate("Form", "Application File"))
        self.label_beam_type.setText(_translate("Form", "Beam Type"))
        self.label_length.setText(_translate("Form", "Length (um)"))
        self.label_time.setText(_translate("Form", "Time (sec)"))



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemMinimapWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(670, 854)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.label_title = QtWidgets.QLabel(self.centralwidget)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_title.setFont(font)
        self.label_title.setObjectName("label_title")
        self.gridLayout.addWidget(self.label_title, 0, 0, 1, 1)
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.tab)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_tile_resolution = QtWidgets.QLabel(self.tab)
        self.label_tile_resolution.setObjectName("label_tile_resolution")
        self.gridLayout_2.addWidget(self.label_tile_resolution, 3, 0, 1, 1)
        self.doubleSpinBox_translation_y = QtWidgets.QDoubleSpinBox(self.tab)
        self.doubleSpinBox_translation_y.setObjectName("doubleSpinBox_translation_y")
        self.gridLayout_2.addWidget(self.doubleSpinBox_translation_y, 15, 1, 1, 1)
        self.doubleSpinBox_tile_grid_size = QtWidgets.QDoubleSpinBox(self.tab)
        self.doubleSpinBox_tile_grid_size.setMaximum(1000000.0)
        self.doubleSpinBox_tile_grid_size.setProperty("value", 400.0)
        self.doubleSpinBox_tile_grid_size.setObjectName("doubleSpinBox_tile_grid_size")
        self.gridLayout_2.addWidget(self.doubleSpinBox_tile_grid_size, 1, 1, 1, 1)
        self.checkBox_tile_autocontrast = QtWidgets.QCheckBox(self.tab)
        self.checkBox_tile_autocontrast.setChecked(True)
        self.checkBox_tile_autocontrast.setObjectName("checkBox_tile_autocontrast")
        self.gridLayout_2.addWidget(self.checkBox_tile_autocontrast, 6, 0, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 18, 0, 1, 2)
        self.spinBox_tile_resolution = QtWidgets.QSpinBox(self.tab)
        self.spinBox_tile_resolution.setMinimum(1)
        self.spinBox_tile_resolution.setMaximum(10000)
        self.spinBox_tile_resolution.setProperty("value", 1024)
        self.spinBox_tile_resolution.setObjectName("spinBox_tile_resolution")
        self.gridLayout_2.addWidget(self.spinBox_tile_resolution, 3, 1, 1, 1)
        self.label_translation_y = QtWidgets.QLabel(self.tab)
        self.label_translation_y.setObjectName("label_translation_y")
        self.gridLayout_2.addWidget(self.label_translation_y, 15, 0, 1, 1)
        self.label_tile_grid_size = QtWidgets.QLabel(self.tab)
        self.label_tile_grid_size.setObjectName("label_tile_grid_size")
        self.gridLayout_2.addWidget(self.label_tile_grid_size, 1, 0, 1, 1)
        self.label_tile_tile_size = QtWidgets.QLabel(self.tab)
        self.label_tile_tile_size.setObjectName("label_tile_tile_size")
        self.gridLayout_2.addWidget(self.label_tile_tile_size, 2, 0, 1, 1)
        self.doubleSpinBox_tile_tile_size = QtWidgets.QDoubleSpinBox(self.tab)
        self.doubleSpinBox_tile_tile_size.setMaximum(100000.0)
        self.doubleSpinBox_tile_tile_size.setProperty("value", 100.0)
        self.doubleSpinBox_tile_tile_size.setObjectName("doubleSpinBox_tile_tile_size")
        self.gridLayout_2.addWidget(self.doubleSpinBox_tile_tile_size, 2, 1, 1, 1)
        self.pushButton_run_tile_collection = QtWidgets.QPushButton(self.tab)
        self.pushButton_run_tile_collection.setObjectName("pushButton_run_tile_collection")
        self.gridLayout_2.addWidget(self.pushButton_run_tile_collection, 10, 0, 1, 2)
        self.label_translation_z = QtWidgets.QLabel(self.tab)
        self.label_translation_z.setObjectName("label_translation_z")
        self.gridLayout_2.addWidget(self.label_translation_z, 16, 0, 1, 1)
        self.label_tile_path = QtWidgets.QLabel(self.tab)
        self.label_tile_path.setObjectName("label_tile_path")
        self.gridLayout_2.addWidget(self.label_tile_path, 7, 0, 1, 1)
        self.lineEdit_tile_path = QtWidgets.QLineEdit(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_tile_path.sizePolicy().hasHeightForWidth())
        self.lineEdit_tile_path.setSizePolicy(sizePolicy)
        self.lineEdit_tile_path.setObjectName("lineEdit_tile_path")
        self.gridLayout_2.addWidget(self.lineEdit_tile_path, 7, 1, 1, 1)
        self.lineEdit_tile_filename = QtWidgets.QLineEdit(self.tab)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_tile_filename.sizePolicy().hasHeightForWidth())
        self.lineEdit_tile_filename.setSizePolicy(sizePolicy)
        self.lineEdit_tile_filename.setObjectName("lineEdit_tile_filename")
        self.gridLayout_2.addWidget(self.lineEdit_tile_filename, 8, 1, 1, 1)
        self.doubleSpinBox_translation_x = QtWidgets.QDoubleSpinBox(self.tab)
        self.doubleSpinBox_translation_x.setObjectName("doubleSpinBox_translation_x")
        self.gridLayout_2.addWidget(self.doubleSpinBox_translation_x, 14, 1, 1, 1)
        self.comboBox_tile_beam_type = QtWidgets.QComboBox(self.tab)
        self.comboBox_tile_beam_type.setObjectName("comboBox_tile_beam_type")
        self.gridLayout_2.addWidget(self.comboBox_tile_beam_type, 0, 1, 1, 1)
        self.label_translation_x = QtWidgets.QLabel(self.tab)
        self.label_translation_x.setObjectName("label_translation_x")
        self.gridLayout_2.addWidget(self.label_translation_x, 14, 0, 1, 1)
        self.label_tile_filename = QtWidgets.QLabel(self.tab)
        self.label_tile_filename.setObjectName("label_tile_filename")
        self.gridLayout_2.addWidget(self.label_tile_filename, 8, 0, 1, 1)
        self.checkBox_options_move_with_translation = QtWidgets.QCheckBox(self.tab)
        self.checkBox_options_move_with_translation.setObjectName("checkBox_options_move_with_translation")
        self.gridLayout_2.addWidget(self.checkBox_options_move_with_translation, 13, 0, 1, 1)
        self.checkBox_tile_autogamma = QtWidgets.QCheckBox(self.tab)
        self.checkBox_tile_autogamma.setChecked(True)
        self.checkBox_tile_autogamma.setObjectName("checkBox_tile_autogamma")
        self.gridLayout_2.addWidget(self.checkBox_tile_autogamma, 6, 1, 1, 1)
        self.checkBox_options_acquire_after_movement = QtWidgets.QCheckBox(self.tab)
        self.checkBox_options_acquire_after_movement.setObjectName("checkBox_options_acquire_after_movement")
        self.gridLayout_2.addWidget(self.checkBox_options_acquire_after_movement, 13, 1, 1, 1)
        self.label_tile_overlap = QtWidgets.QLabel(self.tab)
        self.label_tile_overlap.setObjectName("label_tile_overlap")
        self.gridLayout_2.addWidget(self.label_tile_overlap, 5, 0, 1, 1)
        self.doubleSpinBox_translation_z = QtWidgets.QDoubleSpinBox(self.tab)
        self.doubleSpinBox_translation_z.setObjectName("doubleSpinBox_translation_z")
        self.gridLayout_2.addWidget(self.doubleSpinBox_translation_z, 16, 1, 1, 1)
        self.label_options_move_with_translate_info = QtWidgets.QLabel(self.tab)
        self.label_options_move_with_translate_info.setWordWrap(True)
        self.label_options_move_with_translate_info.setObjectName("label_options_move_with_translate_info")
        self.gridLayout_2.addWidget(self.label_options_move_with_translate_info, 17, 0, 1, 2)
        self.label_tile_beam_type = QtWidgets.QLabel(self.tab)
        self.label_tile_beam_type.setObjectName("label_tile_beam_type")
        self.gridLayout_2.addWidget(self.label_tile_beam_type, 0, 0, 1, 1)
        self.label_options_header = QtWidgets.QLabel(self.tab)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_options_header.setFont(font)
        self.label_options_header.setObjectName("label_options_header")
        self.gridLayout_2.addWidget(self.label_options_header, 11, 0, 1, 2)
        self.label_tile_dwell_time = QtWidgets.QLabel(self.tab)
        self.label_tile_dwell_time.setObjectName("label_tile_dwell_time")
        self.gridLayout_2.addWidget(self.label_tile_dwell_time, 4, 0, 1, 1)
        self.doubleSpinBox_tile_dwell_time = QtWidgets.QDoubleSpinBox(self.tab)
        self.doubleSpinBox_tile_dwell_time.setMaximum(100.0)
        self.doubleSpinBox_tile_dwell_time.setProperty("value", 1.0)
        self.doubleSpinBox_tile_dwell_time.setObjectName("doubleSpinBox_tile_dwell_time")
        self.gridLayout_2.addWidget(self.doubleSpinBox_tile_dwell_time, 4, 1, 1, 1)
        self.tabWidget.addTab(self.tab, "")
        self.tab_4 = QtWidgets.QWidget()
        self.tab_4.setObjectName("tab_4")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.tab_4)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.label_instructions = QtWidgets.QLabel(self.tab_4)
        self.label_instructions.setObjectName("label_instructions")
        self.gridLayout_4.addWidget(self.label_instructions, 5, 0, 1, 2)
        self.label_position_name = QtWidgets.QLabel(self.tab_4)
        self.label_position_name.setObjectName("label_position_name")
        self.gridLayout_4.addWidget(self.label_position_name, 1, 0, 1, 1)
        self.pushButton_update_position = QtWidgets.QPushButton(self.tab_4)
        self.pushButton_update_position.setObjectName("pushButton_update_position")
        self.gridLayout_4.addWidget(self.pushButton_update_position, 2, 0, 1, 1)
        self.label_position_info = QtWidgets.QLabel(self.tab_4)
        self.label_position_info.setObjectName("label_position_info")
        self.gridLayout_4.addWidget(self.label_position_info, 3, 0, 1, 2)
        self.lineEdit_tile_position_name = QtWidgets.QLineEdit(self.tab_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_tile_position_name.sizePolicy().hasHeightForWidth())
        self.lineEdit_tile_position_name.setSizePolicy(sizePolicy)
        self.lineEdit_tile_position_name.setObjectName("lineEdit_tile_position_name")
        self.gridLayout_4.addWidget(self.lineEdit_tile_position_name, 1, 1, 1, 1)
        self.label_pattern_overlay = QtWidgets.QLabel(self.tab_4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_pattern_overlay.setFont(font)
        self.label_pattern_overlay.setObjectName("label_pattern_overlay")
        self.gridLayout_4.addWidget(self.label_pattern_overlay, 6, 0, 1, 2)
        self.label_position_header = QtWidgets.QLabel(self.tab_4)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_position_header.setFont(font)
        self.label_position_header.setObjectName("label_position_header")
        self.gridLayout_4.addWidget(self.label_position_header, 0, 0, 1, 1)
        self.pushButton_move_to_position = QtWidgets.QPushButton(self.tab_4)
        self.pushButton_move_to_position.setObjectName("pushButton_move_to_position")
        self.gridLayout_4.addWidget(self.pushButton_move_to_position, 4, 0, 1, 2)
        self.comboBox_tile_position = QtWidgets.QComboBox(self.tab_4)
        self.comboBox_tile_position.setObjectName("comboBox_tile_position")
        self.gridLayout_4.addWidget(self.comboBox_tile_position, 0, 1, 1, 1)
        self.pushButton_remove_position = QtWidgets.QPushButton(self.tab_4)
        self.pushButton_remove_position.setObjectName("pushButton_remove_position")
        self.gridLayout_4.addWidget(self.pushButton_remove_position, 2, 1, 1, 1)
        spacerItem1 = QtWidgets.QSpacerItem(483, 269, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_4.addItem(spacerItem1, 9, 0, 1, 2)
        self.checkBox_pattern_overlay = QtWidgets.QCheckBox(self.tab_4)
        self.checkBox_pattern_overlay.setObjectName("checkBox_pattern_overlay")
        self.gridLayout_4.addWidget(self.checkBox_pattern_overlay, 7, 0, 1, 1)
        self.comboBox_pattern_overlay = QtWidgets.QComboBox(self.tab_4)
        self.comboBox_pattern_overlay.setObjectName("comboBox_pattern_overlay")
        self.gridLayout_4.addWidget(self.comboBox_pattern_overlay, 7, 1, 1, 1)
        self.tabWidget.addTab(self.tab_4, "")
        self.tab_2 = QtWidgets.QWidget()
        self.tab_2.setObjectName("tab_2")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.tab_2)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.label_correlation_selected_layer = QtWidgets.QLabel(self.tab_2)
        self.label_correlation_selected_layer.setObjectName("label_correlation_selected_layer")
        self.gridLayout_3.addWidget(self.label_correlation_selected_layer, 1, 0, 1, 1)
        self.label_correlation_scale_x = QtWidgets.QLabel(self.tab_2)
        self.label_correlation_scale_x.setObjectName("label_correlation_scale_x")
        self.gridLayout_3.addWidget(self.label_correlation_scale_x, 3, 0, 1, 1)
        self.checkBox_gridbar = QtWidgets.QCheckBox(self.tab_2)
        self.checkBox_gridbar.setObjectName("checkBox_gridbar")
        self.gridLayout_3.addWidget(self.checkBox_gridbar, 5, 0, 1, 1)
        spacerItem2 = QtWidgets.QSpacerItem(483, 261, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_3.addItem(spacerItem2, 9, 0, 1, 2)
        self.doubleSpinBox_correlation_translation_x = QtWidgets.QDoubleSpinBox(self.tab_2)
        self.doubleSpinBox_correlation_translation_x.setDecimals(0)
        self.doubleSpinBox_correlation_translation_x.setMinimum(-100000000.0)
        self.doubleSpinBox_correlation_translation_x.setMaximum(1000000000.0)
        self.doubleSpinBox_correlation_translation_x.setObjectName("doubleSpinBox_correlation_translation_x")
        self.gridLayout_3.addWidget(self.doubleSpinBox_correlation_translation_x, 2, 1, 1, 1)
        self.label_correlation_rotation = QtWidgets.QLabel(self.tab_2)
        self.label_correlation_rotation.setObjectName("label_correlation_rotation")
        self.gridLayout_3.addWidget(self.label_correlation_rotation, 4, 0, 1, 1)
        self.label_correlation_translation_x = QtWidgets.QLabel(self.tab_2)
        self.label_correlation_translation_x.setObjectName("label_correlation_translation_x")
        self.gridLayout_3.addWidget(self.label_correlation_translation_x, 2, 0, 1, 1)
        self.comboBox_correlation_selected_layer = QtWidgets.QComboBox(self.tab_2)
        self.comboBox_correlation_selected_layer.setObjectName("comboBox_correlation_selected_layer")
        self.gridLayout_3.addWidget(self.comboBox_correlation_selected_layer, 1, 1, 1, 2)
        self.doubleSpinBox_correlation_scale_y = QtWidgets.QDoubleSpinBox(self.tab_2)
        self.doubleSpinBox_correlation_scale_y.setDecimals(3)
        self.doubleSpinBox_correlation_scale_y.setMinimum(-100000000.0)
        self.doubleSpinBox_correlation_scale_y.setMaximum(1000000000.0)
        self.doubleSpinBox_correlation_scale_y.setProperty("value", 1.0)
        self.doubleSpinBox_correlation_scale_y.setObjectName("doubleSpinBox_correlation_scale_y")
        self.gridLayout_3.addWidget(self.doubleSpinBox_correlation_scale_y, 3, 2, 1, 1)
        self.label_gb_width = QtWidgets.QLabel(self.tab_2)
        self.label_gb_width.setObjectName("label_gb_width")
        self.gridLayout_3.addWidget(self.label_gb_width, 6, 0, 1, 1)
        self.doubleSpinBox_correlation_translation_y = QtWidgets.QDoubleSpinBox(self.tab_2)
        self.doubleSpinBox_correlation_translation_y.setDecimals(0)
        self.doubleSpinBox_correlation_translation_y.setMinimum(-100000000.0)
        self.doubleSpinBox_correlation_translation_y.setMaximum(1000000000.0)
        self.doubleSpinBox_correlation_translation_y.setObjectName("doubleSpinBox_correlation_translation_y")
        self.gridLayout_3.addWidget(self.doubleSpinBox_correlation_translation_y, 2, 2, 1, 1)
        self.doubleSpinBox_correlation_scale_x = QtWidgets.QDoubleSpinBox(self.tab_2)
        self.doubleSpinBox_correlation_scale_x.setDecimals(3)
        self.doubleSpinBox_correlation_scale_x.setMinimum(-100000000.0)
        self.doubleSpinBox_correlation_scale_x.setMaximum(1000000000.0)
        self.doubleSpinBox_correlation_scale_x.setProperty("value", 1.0)
        self.doubleSpinBox_correlation_scale_x.setObjectName("doubleSpinBox_correlation_scale_x")
        self.gridLayout_3.addWidget(self.doubleSpinBox_correlation_scale_x, 3, 1, 1, 1)
        self.pushButton_update_correlation_image = QtWidgets.QPushButton(self.tab_2)
        self.pushButton_update_correlation_image.setObjectName("pushButton_update_correlation_image")
        self.gridLayout_3.addWidget(self.pushButton_update_correlation_image, 8, 1, 1, 2)
        self.doubleSpinBox_correlation_rotation = QtWidgets.QDoubleSpinBox(self.tab_2)
        self.doubleSpinBox_correlation_rotation.setDecimals(2)
        self.doubleSpinBox_correlation_rotation.setMinimum(-360.0)
        self.doubleSpinBox_correlation_rotation.setMaximum(360.0)
        self.doubleSpinBox_correlation_rotation.setObjectName("doubleSpinBox_correlation_rotation")
        self.gridLayout_3.addWidget(self.doubleSpinBox_correlation_rotation, 4, 1, 1, 2)
        self.label_gb_spacing = QtWidgets.QLabel(self.tab_2)
        self.label_gb_spacing.setObjectName("label_gb_spacing")
        self.gridLayout_3.addWidget(self.label_gb_spacing, 7, 0, 1, 1)
        self.doubleSpinBox_gb_width = QtWidgets.QDoubleSpinBox(self.tab_2)
        self.doubleSpinBox_gb_width.setMaximum(10000.0)
        self.doubleSpinBox_gb_width.setObjectName("doubleSpinBox_gb_width")
        self.gridLayout_3.addWidget(self.doubleSpinBox_gb_width, 6, 1, 1, 1)
        self.doubleSpinBox_gb_spacing = QtWidgets.QDoubleSpinBox(self.tab_2)
        self.doubleSpinBox_gb_spacing.setMaximum(10000.0)
        self.doubleSpinBox_gb_spacing.setObjectName("doubleSpinBox_gb_spacing")
        self.gridLayout_3.addWidget(self.doubleSpinBox_gb_spacing, 7, 1, 1, 1)
        self.tabWidget.addTab(self.tab_2, "")
        self.gridLayout.addWidget(self.tabWidget, 1, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 670, 22))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionLoad_Image = QtWidgets.QAction(MainWindow)
        self.actionLoad_Image.setObjectName("actionLoad_Image")
        self.actionLoad_Correlation_Image = QtWidgets.QAction(MainWindow)
        self.actionLoad_Correlation_Image.setObjectName("actionLoad_Correlation_Image")
        self.actionSave_Positions = QtWidgets.QAction(MainWindow)
        self.actionSave_Positions.setObjectName("actionSave_Positions")
        self.actionLoad_Positions = QtWidgets.QAction(MainWindow)
        self.actionLoad_Positions.setObjectName("actionLoad_Positions")
        self.menuFile.addAction(self.actionLoad_Image)
        self.menuFile.addAction(self.actionLoad_Correlation_Image)
        self.menuFile.addAction(self.actionSave_Positions)
        self.menuFile.addAction(self.actionLoad_Positions)
        self.menubar.addAction(self.menuFile.menuAction())

        self.retranslateUi(MainWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label_title.setText(_translate("MainWindow", "Minimap "))
        self.label_tile_resolution.setText(_translate("MainWindow", "Resolution (px)"))
        self.checkBox_tile_autocontrast.setText(_translate("MainWindow", "AutoContrast"))
        self.label_translation_y.setText(_translate("MainWindow", "Translation Y (mm)"))
        self.label_tile_grid_size.setText(_translate("MainWindow", "Grid Size (um)"))
        self.label_tile_tile_size.setText(_translate("MainWindow", "Tile Size (um)"))
        self.pushButton_run_tile_collection.setText(_translate("MainWindow", "Run Tile Collection"))
        self.label_translation_z.setText(_translate("MainWindow", "Translation Z (mm)"))
        self.label_tile_path.setText(_translate("MainWindow", "Path"))
        self.lineEdit_tile_filename.setText(_translate("MainWindow", "overview-image"))
        self.label_translation_x.setText(_translate("MainWindow", "Translation X (mm)"))
        self.label_tile_filename.setText(_translate("MainWindow", "Filename"))
        self.checkBox_options_move_with_translation.setText(_translate("MainWindow", "Move with Translated Position"))
        self.checkBox_tile_autogamma.setText(_translate("MainWindow", "AutoGamma"))
        self.checkBox_options_acquire_after_movement.setText(_translate("MainWindow", "Acquire Image After Movement"))
        self.label_tile_overlap.setText(_translate("MainWindow", "Overlap (%)"))
        self.label_options_move_with_translate_info.setText(_translate("MainWindow", "Move with Translated Position allows you to click on a position in the minimap, and move while translated away from the beams, e.g. under an integrated light microscope."))
        self.label_tile_beam_type.setText(_translate("MainWindow", "Beam Type"))
        self.label_options_header.setText(_translate("MainWindow", "Movement Options"))
        self.label_tile_dwell_time.setText(_translate("MainWindow", "Dwell Time (us)"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("MainWindow", "Collection"))
        self.label_instructions.setText(_translate("MainWindow", "Please take or load an overview image."))
        self.label_position_name.setText(_translate("MainWindow", "Position Name"))
        self.pushButton_update_position.setText(_translate("MainWindow", "Update Position"))
        self.label_position_info.setText(_translate("MainWindow", "No Positions saved."))
        self.label_pattern_overlay.setText(_translate("MainWindow", "Pattern Overlay"))
        self.label_position_header.setText(_translate("MainWindow", "Saved Positions"))
        self.pushButton_move_to_position.setText(_translate("MainWindow", "Move to Position"))
        self.pushButton_remove_position.setText(_translate("MainWindow", "Remove Position"))
        self.checkBox_pattern_overlay.setText(_translate("MainWindow", "Display Pattern"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_4), _translate("MainWindow", "Positions"))
        self.label_correlation_selected_layer.setText(_translate("MainWindow", "Selected Layer"))
        self.label_correlation_scale_x.setText(_translate("MainWindow", "Scale x, y (%)"))
        self.checkBox_gridbar.setText(_translate("MainWindow", "Overlay Gridbar"))
        self.label_correlation_rotation.setText(_translate("MainWindow", "Rotation (deg)"))
        self.label_correlation_translation_x.setText(_translate("MainWindow", "Translation x, y (px)"))
        self.label_gb_width.setText(_translate("MainWindow", "Gridbar Width (um)"))
        self.pushButton_update_correlation_image.setText(_translate("MainWindow", "Update Correlation Image"))
        self.label_gb_spacing.setText(_translate("MainWindow", "Gridbar Spacing (um)"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab_2), _translate("MainWindow", "Correlation"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.actionLoad_Image.setText(_translate("MainWindow", "Load Image"))
        self.actionLoad_Correlation_Image.setText(_translate("MainWindow", "Load Correlation Image"))
        self.actionSave_Positions.setText(_translate("MainWindow", "Save Positions"))
        self.actionLoad_Positions.setText(_translate("MainWindow", "Load Positions"))
